#include "apr_file_io.h"
#include "apr_file_info.h"
#include "apr_network_io.h"
#include "apr_errno.h"
#include "apr_general.h"
#include "apr_strings.h"
#include "apr_poll.h"
#include "apr_lib.h"
#include "http_protocol.h"

#include "pcre.h"
#include <assert.h>

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

#if 0
#include "ldo.h"
#include "lfunc.h"
#include "lmem.h"
#include "lobject.h"
#include "lopcodes.h"
#include "lstring.h"
#include "lundump.h"
#endif
#include "storage.h"

#define OVECCOUNT 6    /* should be a 2 * 3 */
const char* compile_lhtml(apr_pool_t *pool, const char* content, int *step)
{
	int pc;
	pcre *re;
	const char *error;
	int erroffset;
	const char *pattern[] ={"<\\?lua(.*?)\\?>", "<%\\s*=(.*?)%>", "<%(.*?)%>", NULL };
	int ovector[OVECCOUNT];

	assert(content);

    re = pcre_compile(
		pattern[*step],       /* the pattern */
		PCRE_DOTALL,          /* default options */
		&error,               /* for error message */
		&erroffset,           /* for error offset */
		NULL);                /* use default character tables */

	/* Compilation failed: print the error message and exit */
    assert(re!=NULL);
	if (re == NULL)
	{
		*step=-1;
		return NULL;
	}
/*
When a match has been successful, information about captured substrings is returned in pairs of integers,
starting at the beginning of ovector, and continuing up to two-thirds of its length at the most.

ƥɹ󣬲ӴϢڷصζ,Ŀʼ2/3ĳ.
The first element of a pair is set to the offset of the first character in a substring, and the second is
set to the offset of the first character after the end of a substring. The first pair, ovector[0] and ovector[1],
identify the portion of the subject string matched by the entire pattern. 
еĵһԪӴеһַλãڶӴĵһַλãһԣʾȫƥӴλ

The next pair is used for the first capturing subpattern, and so on. The value returned by pcre_exec() 
is the number of pairs that have been set. If there are no capturing subpatterns, the return value from
a successful match is 1, indicating that just the first pair of offsets has been set.
Ķӵһƥģʽ
*/
	if(*step==0)  //<?lua ?>
	{
		while( (pc = pcre_exec(
			re,                   /* the compiled pattern */
			NULL,                 /* no extra data - we didn't study the pattern */
			content,              /* the subject string */
			(int)strlen(content), /* the length of the subject */
			0,			          /* start at offset 0 in the subject */
			0,                    /* default options */
			ovector,              /* output vector for substring information */
			OVECCOUNT)           /* number of elements in the output vector */
			)==2 )
		{
			content = apr_psprintf(pool,
				"%.*s<%%%.*s%%>%s",
				ovector[0],
				content,
				ovector[3] - ovector[2],
				content+ovector[2],
				content + ovector[1]);
		}
	}else if(*step==1)
	{
		while( (pc = pcre_exec(
			re,                   /* the compiled pattern */
			NULL,                 /* no extra data - we didn't study the pattern */
			content,              /* the subject string */
			(int)strlen(content), /* the length of the subject */
			0,			          /* start at offset 0 in the subject */
			0,                    /* default options */
			ovector,              /* output vector for substring information */
			OVECCOUNT)           /* number of elements in the output vector */
			)==2 )
		{
			content = apr_psprintf(pool,
				"%.*s<%% ap.rputs(%.*s) %%>%s",
				ovector[0],
				content,
				ovector[3] - ovector[2],
				content+ovector[2],
				content + ovector[1]);
		}
	}else if(*step==2)
	{
		content = apr_psprintf(pool,"ap.rputs [[\n%s\n]]",content);
		while( (pc = pcre_exec(
			re,                   /* the compiled pattern */
			NULL,                 /* no extra data - we didn't study the pattern */
			content,              /* the subject string */
			(int)strlen(content), /* the length of the subject */
			0,			          /* start at offset 0 in the subject */
			0,                    /* default options */
			ovector,              /* output vector for substring information */
			OVECCOUNT)           /* number of elements in the output vector */
			)==2 )
		{
			content = apr_psprintf(pool,
				"%.*s]]\n%.*s\nap.rputs [[%s",
				ovector[0],
				content,
				ovector[3] - ovector[2],
				content+ovector[2],
				content + ovector[1]);
		}
	}

	*step += 1;
	return content;
}
/************************************************************************/
/* lhtmlLUAű                                             */
/* ɾ lua ܲʲôõluaδ   */
/************************************************************************/

