//Apache Header
#include "httpd.h"          /* defines request_rec */
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

//Mod_lua Header
#include "storage.h"
#include "mod_lua.h"
#include "apache2_lib.h"

#include <assert.h>
#include <stdarg.h>
#include <math.h>

#if AP_SERVER_MAJORVERSION_NUMBER!=2 || (AP_SERVER_MINORVERSION_NUMBER!=0 && AP_SERVER_MINORVERSION_NUMBER!=2)
#error Sorry, Only support Apache2
#endif

#define LUA_HANDLE "lua-handle"
#define LUA_SCRIPT "lua-script"
#define LUA_HTML   "lua-html"

#define LUA2REC(L) ( lua_isuserdata(L,4) ? *(request_rec **)lua_touserdata (L, 4) : NULL )
/************************************************************************/
/*  ڲк,                                                                    */
/************************************************************************/
#if defined(_DEBUG) || defined(DEBUG)
void stack_dump (lua_State *L) {
	int i;
	int top = lua_gettop(L);
	printf("Stack Length:%d ",top);
	for (i = 1; i <= top; i++) 
	{   /* repeat for each level */
		int t = lua_type(L, i);
		switch (t) {
		  case LUA_TNIL:
			  printf("nil");
			  break;
		  case LUA_TBOOLEAN:  /* booleans */
			  printf("boolean:%s", (lua_toboolean(L, i) ? "TRUE" : "FALSE"));
			  break;
		  case LUA_TLIGHTUSERDATA:
			  printf("lightuserdata:%p",lua_touserdata(L,i));
			  break;
		  case LUA_TNUMBER:  /* numbers */
			  printf("number:%g", lua_tonumber(L, i));
			  break;
		  case LUA_TSTRING:  /* strings */
			  printf("`string:%s'", lua_tostring(L, i));
			  break;
		  case LUA_TTABLE:
		  case LUA_TFUNCTION:
		  case LUA_TUSERDATA:
		  case LUA_TTHREAD:
		  default:  /* other values */
			  printf("%s:%p", lua_typename(L, t),lua_topointer(L,i) );
			  break;
		}
		printf("  ");  /* put a separator */
	}
	printf("\n");  /* end the listing */
}
#endif

//////////////////////////////////////////////////////////////////////////
//Call in C, stack  ͨlua òһ
//copy from lua.c, small modify
//////////////////////////////////////////////////////////////////////////
static int ap2_alert (lua_State *L) 
{
	const char *p = lua_tostring(L,-1);
	request_rec* r = LUA2REC(L);

	if(r)
	{
		r->status = HTTP_INTERNAL_SERVER_ERROR;
		ap_rprintf(r,"<TEXTAREA COLS=80 ROWS=25>%s</TEXTAREA>",p);
		ap_rflush(r);
	}
	lua_pop(L, 1);
	return 0; /* number of results */
}

static int report (lua_State *L, int status) {
	if (status && !lua_isnil(L, -1)) 
	{
		ap2_alert(L);
		status = HTTP_INTERNAL_SERVER_ERROR;
	}

	//try to get lua return value
	if (lua_isnumber(L,-1))
		status = lua_tointeger (L, -1);

	return status;
}


static int traceback (lua_State *L) 
{
	lua_getfield(L, LUA_GLOBALSINDEX, "debug");
	if (!lua_istable(L, -1)) {
		lua_pop(L, 1);
		return 1;
	}
	lua_getfield(L, -1, "traceback");
	if (!lua_isfunction(L, -1)) {
		lua_pop(L, 2);
		return 1;
	}
	// pass error message
	lua_pushvalue(L, 1);  
	// skip this function and traceback
	lua_pushinteger(L, 2);  
	// call debug.traceback 
	lua_call(L, 2, 1);  
	return 1;
}

static int docall (lua_State *L, int narg, int clear) {
	int status;
	// function index 
	int base = lua_gettop(L) - narg;  
	// push traceback function
	lua_pushcfunction(L, traceback);  
	// put it under chunk and args 
	lua_insert(L, base);  
	status = lua_pcall(L, narg, (clear ? 0 : LUA_MULTRET), base);

	// remove traceback function
	lua_remove(L, base);  

	// force a complete garbage collection in case of errors 
	if (status != 0) 
		lua_gc(L, LUA_GCCOLLECT, 0);
	return status;
}

static int handle_luascript (lua_State *L, const char* script, const char* title) {
	int status = 0;

	if (script == NULL) 
		return 0;  
	/* status OK */
	else if (script[0] == '@')
	{
		status = luaL_loadfile(L, script+1) || docall(L, 0, 0);
		//status = luaL_dofile(L, script+1);
	}
	else
	{
		status = luaL_loadbuffer(L, script, strlen(script), title) || docall(L, 0, 0);
		//status = luaL_dostring(L, script);
	}

	return report(L,status);
}

//call lua function
/************************************************************************/
/* ʹCvarargװLuaĵáǵķװĺcall_lua_va*/
/* ܱõĺΪһڶһͽ͵ */
/* ַһ淵ؽıָб                       */
/************************************************************************/
static int call_lua_va (request_rec* r, lua_State *L, const char *func, const char *sig, ...) {
	va_list vl;
	int narg, nres;	/* number of arguments and results */

	va_start(vl, sig);
	lua_getglobal(L, func);	/* get function */

	/* push arguments */
	narg = 0;
	while (*sig) {	/* push arguments */
		switch (*sig++) {

		case 'd':	/* double argument */
			lua_pushnumber(L, va_arg(vl, double));
			break;

		case 'i':	/* int argument */
			lua_pushnumber(L, va_arg(vl, int));
			break;

		case 's':	/* string argument */
			lua_pushstring(L, va_arg(vl, char *));
			break;

		case '>':
			goto endwhile;

		default:
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "call_lua_va invalid option (%c)", *(sig - 1));
			return -1;
		}
		narg++;
		luaL_checkstack(L, 1, "too many arguments");
	}
endwhile:

	/* do the call */
	nres = strlen(sig);	/* number of expected results */
	if (lua_pcall(L, narg, nres, 0) != 0)	/* do the call */
	{
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "error running function `%s': %s", func, lua_tostring(L, -1));
		return -1;
	}


	/* retrieve results */
	nres = -nres;		/* stack index of first result */
	while (*sig) {	/* get results */
		switch (*sig++) {

		case 'd':	/* double result */
			if (!lua_isnumber(L, nres))
			{
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "wrong result type");
				return -1;
			}
			*va_arg(vl, double *) = lua_tonumber(L, nres);
			break;

		case 'i':	/* int result */
			if (!lua_isnumber(L, nres))
			{
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "wrong result type");
				return -1;
			}
			*va_arg(vl, int *) = (int)lua_tonumber(L, nres);
			break;

		case 's':	/* string result */
			if (!lua_isstring(L, nres))
			{
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "wrong result type");
				return -1;
			}
			*va_arg(vl, const char **) = lua_tostring(L, nres);
			break;

		default:
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "invalid option (%c)", *(sig - 1));
			return -1;
		}
		nres++;
	}
	va_end(vl);
	return 0;
}

/************************************************************************/
/* ߳ģʽLUAͬһ̿ռ乤Ҫмѭ    */
/************************************************************************/
/************************************************************************/
/* 0.2 汾Ժ󣬲LUA    */
/************************************************************************/

/***  lua VM лȡе lua VM ***/
static lua_State* ml_openlua(request_rec *r)
{
	apr_status_t status = APR_EINTR;
	lua_State *L = NULL;

	LUASrvConfigRec *sc = ThisSrvConfig(VHOST(r));
	
	/*
	APR_EINTR the blocking was interrupted (try again)
	APR_EOF if the queue has been terminated
	APR_SUCCESS on a successfull pop 
	*/
#ifdef APR_HAS_THREADS
	if(sc->is_threaded_mpm)
	{
		while(status == APR_EINTR)
			status = apr_queue_pop(sc->process_queue, &L);
		if(status == APR_EOF)
			return NULL;
		assert(status == APR_SUCCESS);
	}else
		L = sc->L;
#else
	L = sc->L;
#endif

	assert(L);

	//LUA VM Bind request_r
	ap2_set_request_rec (L, r);

	status = handle_luascript(L,SrvConfig4Request(r)->http_init,"=HTTP_INIT");
	if (status != OK)
	{
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"LUA can't do HTTP_INIT handle %s for %s",SrvConfig4Request(r)->main_script,r->uri);
	}

	return L;
}

/***  lua VM ӵг ***/
static apr_status_t ml_closelua(request_rec *r, lua_State *L)
{
	apr_status_t status = APR_EINTR;
	LUASrvConfigRec *sc = ThisSrvConfig(VHOST(r));
#ifdef APR_HAS_THREADS
	if(sc->is_threaded_mpm)
	{
		while(status == APR_EINTR)
			status = apr_queue_push(sc->process_queue, L);
		return status;
	}
#endif
	return APR_SUCCESS;
	// APR_EOF or APR_SUCCESS;
}


/************************************************************************/
/* lua_handler                                                          */
/************************************************************************/
const char* compile_lhtml(apr_pool_t *pool, const char* content, int *step);
int comiple_lpage(const char* content, const char* saveto, request_rec *r);
//ҪϲĴʽ
static int mod_lua_dostring(request_rec *r, const char* content)
{
	int rc = APR_SUCCESS;
	lua_State* L = ml_openlua(r);

	assert(L!=NULL);
	rc = luaL_loadbuffer(L, content, strlen(content), NULL) || docall(L, 0, 0);
	//rc = luaL_dostring(L,content);
	rc = report(L, rc);
	if (rc == OK)
		rc = ap_discard_request_body (r);

	ml_closelua(r,L);

	return rc;
}

static int mod_lua_dofile (request_rec *r, const char *name) 
{
	int rc = APR_SUCCESS;
	lua_State* L = ml_openlua(r);

	assert(L!=NULL);
	rc = luaL_loadfile(L, name) || docall(L, 0, 0);
	//rc = luaL_dofile(L,name);
	rc = report(L, rc);
	if (rc == OK)
		rc = ap_discard_request_body (r);

	ml_closelua(r,L);

	return rc;
}

//handler
typedef enum {e_script=0, e_html, e_handle} handler_t;

static apr_status_t lua_handler(request_rec *r)
{
	int status;
	const char *val;
	apr_table_t *e;
	handler_t handler;

	if(strcmp(r->handler, LUA_HANDLE)==0)
		handler = e_handle;
	else if(strcmp(r->handler, LUA_SCRIPT)==0)
		handler = e_script;
	else if(strcmp(r->handler, LUA_HTML)==0)
		handler = e_html;
	else
		return DECLINED;

	//ȱʡContent-Type,л޸
	ap_set_content_type(r,"text/html");

    //׼CGIݵĻ
	e = r->subprocess_env;
	ap_add_common_vars(r);
	ap_add_cgi_vars(r);
	apr_table_setn(e, "UNMAPPED_REMOTE_USER", "REMOTE_USER");
	val = apr_table_get(e, "HTTPS");
	if (val && strcmp(val, "on"))
		apr_table_setn(e, "SERVER_PORT_SECURE", "1");
	else
		apr_table_setn(e, "SERVER_PORT_SECURE", "0");
	apr_table_setn(e, "URL", r->uri);

	if(handler==e_handle)
	{
		ml_directory *ml_dir = (ml_directory *) ap_get_module_config (r->per_dir_config, &lua_module);

		if(ml_dir->handlers!=NULL)
		{
			int rc = APR_SUCCESS;
			lua_State* L = ml_openlua(r);
			assert(L!=NULL);
			rc = handle_luascript(L,ml_dir->handlers,"=LUA_HANDLE");
			ml_closelua(r,L);
			if (rc == OK)
				rc = ap_discard_request_body (r);
			else
			{
				ap_log_rerror(APLOG_MARK, APLOG_ERR, rc, r,
					"LUA can't handle %s with %s",r->uri,ml_dir->handlers);
				return HTTP_INTERNAL_SERVER_ERROR;
			}
			return OK;
		}
		else
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"LUA handle not set to %s",r->uri);
		return HTTP_INTERNAL_SERVER_ERROR;
	}

	//ȫmain_script,handlee_handle
	if(SrvConfig4Request(r)->main_script!=NULL)
	{
		int rc = APR_SUCCESS;
		lua_State* L = ml_openlua(r);
		assert(L!=NULL);
		rc = handle_luascript(L,SrvConfig4Request(r)->main_script,"=LUA_MAIN");
		ml_closelua(r,L);
		if (rc == OK)
			rc = ap_discard_request_body (r);
		else
		{
			ap_log_rerror(APLOG_MARK, APLOG_ERR, rc, r,
				"LUA can't handle %s with %s",r->uri,SrvConfig4Request(r)->main_script);
			return HTTP_INTERNAL_SERVER_ERROR;
		}
		return OK;
	}

	if(handler == e_script)
	{
		if(r->finfo.filetype!=APR_REG) 
			return HTTP_NOT_FOUND;

		if (r->finfo.size > 1024*1204 )
		{
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
				"%s too large(%I64d) for process",r->filename,r->finfo.size);

			return HTTP_INTERNAL_SERVER_ERROR;
		}
		status = mod_lua_dofile(r, r->filename);
		if (status != 0)
		{
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
			"LUA can't handle %s(%d)",r->filename,status);
		}

		return status;
	}


	if(handler==e_html)
	{
		char *pathtmp;
		apr_file_t *file;

		apr_status_t rc;
		apr_size_t size;
		apr_finfo_t finfo;


		char *content;
		char *filename;
		char *pos;
		char errbuff[MAX_STRING_LEN];
		int step;

		//ǷҪ±
		pos = strrchr(r->filename,'/');
		assert(pos);

		//extend printf Format
		pathtmp = apr_psprintf(r->pool, "%.*s.tmp", (pos - r->filename+1), r->filename);
		filename = apr_psprintf(r->pool, "%s%s.lua",pathtmp,pos);

		rc = apr_stat(&finfo, filename, APR_FINFO_MTIME, r->pool);

		if(APR_STATUS_IS_ENOENT(rc))
		{	//ļ
			//Ŀ¼Ƿ
			rc = apr_stat(&finfo, pathtmp, APR_FINFO_TYPE, r->pool);
			if(rc!=APR_SUCCESS &&  APR_STATUS_IS_ENOENT(rc))
			{
				//һµĿ¼
				rc = apr_dir_make(pathtmp,APR_OS_DEFAULT,r->pool);
				if(rc!=APR_SUCCESS)
				{
					ap_log_rerror(APLOG_MARK, APLOG_ERR, rc, r,
						"Error compile %s because %s",r->filename,apr_strerror(rc,errbuff,MAX_STRING_LEN));
					return HTTP_INTERNAL_SERVER_ERROR;
				}
			}else if(finfo.filetype != APR_DIR)
			{
				ap_log_rerror(APLOG_MARK, APLOG_ERR, rc, r,
					"%s not a directory",pathtmp);

				return HTTP_INTERNAL_SERVER_ERROR;
			}
			finfo.mtime = 0;
		}
        if(rc!=APR_SUCCESS)
		{
			ap_log_rerror(APLOG_MARK, APLOG_ERR, rc, r,
				"Error compile %s because %s",r->filename,apr_strerror(rc,errbuff,MAX_STRING_LEN));

			return HTTP_INTERNAL_SERVER_ERROR;
		}

		if(finfo.mtime < r->finfo.mtime)
		{
			//ļѸ£±

			//Read File
			rc = apr_file_open(&file, r->filename,APR_READ, APR_OS_DEFAULT, r->pool);
			if (rc!=APR_SUCCESS) {
				ap_rprintf(r,"Error compile %s because %s",
					r->filename,
					apr_strerror(rc,errbuff,MAX_STRING_LEN) );

				ap_log_rerror(APLOG_MARK, APLOG_ERR,rc, r,
					"can't open %s",r->filename);

				return HTTP_INTERNAL_SERVER_ERROR;
			}

			content = apr_palloc(r->pool,(int)r->finfo.size+1);
			rc = apr_file_read_full(file,content,(apr_size_t)r->finfo.size,&size);
			assert(rc==APR_SUCCESS && size==r->finfo.size);
			content[r->finfo.size] = '\0';
			apr_file_close(file);

			//Compile lhtml
			step = 0;
			while(step<3)
				content = (char *)compile_lhtml(r->pool,content,&step);

			rc = mod_lua_dostring(r, content);
			if (rc!=0)
			{
				ap_log_rerror(APLOG_MARK, APLOG_ERR,rc, r,
					"LUA can't do %s",r->filename);

				ap_rprintf(r,"\n<TEXTAREA COLS=80 ROWS=25>%s</TEXTAREA>\n",content);
				ap_rflush(r);
				return rc;
			}
			
			rc = apr_file_open(&file,filename, APR_CREATE|APR_WRITE|APR_TRUNCATE|APR_BINARY, APR_OS_DEFAULT, r->pool);
			if(rc==APR_SUCCESS)
			{
				size_t s = strlen(content);
				rc = apr_file_write(file, content, &s);
				assert( rc==APR_SUCCESS && strlen(content)==s);
				if(rc==APR_SUCCESS)
					rc = apr_file_close(file);
			}
			if(rc!=APR_SUCCESS)
				ap_log_rerror(APLOG_MARK, APLOG_ERR,rc, r, "save %s error,can't do %s",filename,content);

			return OK;
		}
		rc = mod_lua_dofile(r, filename);
		if( rc!= 0) 
		{
			ap_log_rerror(APLOG_MARK, APLOG_ERR,0, r, "LUA can't do %s(%d)",filename,rc);
		}
		return rc;
	}

	return HTTP_INTERNAL_SERVER_ERROR;
}


static apr_status_t lua_access(request_rec *r)
{
	apr_status_t status = 0;
	lua_State *L = NULL;
	LUASrvConfigRec *sc = ThisSrvConfig(r->server);
	if(sc->access_check==NULL)
		return DECLINED;

	L = ml_openlua(r);
	if(call_lua_va(r,L,sc->access_check,"s>i",r->uri,&status)==-1)
		status = HTTP_INTERNAL_SERVER_ERROR;
	ml_closelua(r,L);
	return status;
};

static int lua_check_user_id(request_rec *r)
{
	apr_status_t status = 0;
	char *user = NULL;
	char *auth_type = NULL;

	lua_State *L = NULL;
	LUASrvConfigRec *sc = ThisSrvConfig(r->server);
	if(sc->access_check==NULL)
		return DECLINED;

	L = ml_openlua(r);
	if(call_lua_va(r,L,sc->check_user_id,"s>ss",r->uri,&user,&auth_type)==-1)
		status = HTTP_INTERNAL_SERVER_ERROR;
	r->user = apr_pstrdup(r->pool,user);
	r->ap_auth_type = apr_pstrdup(r->pool,auth_type);
	ml_closelua(r,L);
	return status;
}



static apr_status_t lua_child_clear(lua_State *L)
{
	//Ӧöε
	assert(L);
	lua_close(L);
	return APR_SUCCESS;
}

static void lua_child_init(apr_pool_t *p, server_rec *s)
{
	apr_status_t status = APR_SUCCESS;
	LUASrvConfigRec *sc = ThisSrvConfig(s);

#ifdef APR_HAS_THREADS
	int i = 0;
	int len = sc->is_threaded_mpm ? sc->process_queue_length : 1;
	//ʼ lua Pool
	for(i = 0; i < len; i++)
	{
#endif
		lua_State *L = lua_open(); 


		luaL_openlibs(L);

		if(!luaopen_ap2(L))   //Open {"apache", luaopen_ap2}
		{
			ap_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, status, s,
				"failed to initialize LUA (luaopen_ap2)");
		}

		status = handle_luascript(L,sc->init_script, "=LUA_INIT");
		if (status != 0) 
		{
			ap_log_error(APLOG_MARK, APLOG_CRIT | APLOG_NOERRNO, status, s,
				"failed to initialize LUA script: %s", sc->init_script );
		}

		/* redefine error handler */
		lua_pushcfunction(L, ap2_alert);
		lua_setglobal(L, "_ALERT");

		apr_pool_cleanup_register(p, L, (apr_status_t (*)(void *))lua_child_clear, apr_pool_cleanup_null);


#ifdef APR_HAS_THREADS 
		if (sc->is_threaded_mpm)
			apr_queue_push(sc->process_queue,L);
		else
			sc->L = L;
	}
#else
		sc->L = L;
#endif
#ifdef APR_HAS_THREADS
	if (sc->is_threaded_mpm)
		assert(len == apr_queue_size(sc->process_queue));
#endif

}

#define MOD_LUA_STRING_VERSION  "mod_lua/0.5"

static int modlua_prepare(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp)
{
	return APR_SUCCESS;
}

static int modlua_startup(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	ap_add_version_component(p, MOD_LUA_STRING_VERSION);
	ap_add_version_component(p, LUA_VERSION);
	return storage_init_Module(p,plog,ptemp,s);
}

void modlua_init_Child(apr_pool_t *p, server_rec *s)
{
	storage_init_Child(p,s);
	lua_child_init(p,s);
}

/*
** Create per-directory configuration structure with default values.
** Initiate handlers structure.
*/
static void *create_ml_directory (apr_pool_t *p, char *dirspec) {
	ml_directory *dir = (ml_directory *)apr_pcalloc (p, sizeof(ml_directory));
	(void)dirspec; /* avoid "unused parameter" warning */
	dir->handlers = NULL;
	return (void *) dir;
}


/*
** Merge two directory configuration structures.
** Copy all handlers from both structures (no check to avoid duplications).
*/
static void *merge_ml_directory (apr_pool_t *p, void *parent_conf, void *newloc_conf) {
	ml_directory *dir = (ml_directory *) apr_pcalloc (p, sizeof(ml_directory));
	ml_directory *d1 = (ml_directory *) parent_conf;
	ml_directory *d2 = (ml_directory *) newloc_conf;

	dir->handlers = NULL;
	if(d1->handlers)
		dir->handlers=d1->handlers;
	if(dir->handlers==NULL && d2->handlers)
		dir->handlers=d2->handlers;

	return (void *)dir;
}

// ======================================
/*
*  the table of configuration directives we provide
*/

const char lua_valid_mutex_string[] =
"Valid LUA Session Mutex mechanisms are: `none', `default'"
#if APR_HAS_FLOCK_SERIALIZE
", `flock:/path/to/file'"
#endif
#if APR_HAS_FCNTL_SERIALIZE
", `fcntl:/path/to/file'"
#endif
#if APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)
", `sysvsem'"
#endif
#if APR_HAS_POSIXSEM_SERIALIZE
", `posixsem'"
#endif
#if APR_HAS_PROC_PTHREAD_SERIALIZE
", `pthread'"
#endif
#if APR_HAS_FLOCK_SERIALIZE || APR_HAS_FCNTL_SERIALIZE
", `file:/path/to/file'"
#endif
#if (APR_HAS_SYSVSEM_SERIALIZE && !defined(PERCHILD_MPM)) || APR_HAS_POSIXSEM_SERIALIZE
", `sem'"
#endif
" ";


static const command_rec lua_cmds[] = {
	/*
	* Global (main-server) context configuration directives
	*/
	AP_INIT_TAKE1("LUA_Mutex", lua_cmd_Mutex, 
		NULL, RSRC_CONF|OR_AUTHCFG, lua_valid_mutex_string),

	AP_INIT_TAKE1("LUA_Storage", lua_cmd_Storage,
		NULL, RSRC_CONF,
		"LUA storage "
		"(`none', `dbm:/path/to/file')"),

	AP_INIT_TAKE1("LUA_StorageTimeout", lua_cmd_StorageTimeout,
		NULL, RSRC_CONF,
		"LUA Storage object lifetime "
		"(`N' - number of seconds)"),

	AP_INIT_FLAG("LUA_Release", lua_cmd_Release,
		NULL, RSRC_CONF,
		"LUA Release Switch "
		"(on or off)"),

	AP_INIT_TAKE1("LUA_ACCESS_CHECK", lua_cmd_Access,
		NULL, ACCESS_CONF|RSRC_CONF|OR_AUTHCFG,
		"LUA Request Access Check Handle Function "
		"(`lua function has been defined', `int access_check(uri)')"),

	AP_INIT_TAKE1("LUA_CHECK_USER_ID", lua_cmd_Check_user_id,
		NULL, ACCESS_CONF|RSRC_CONF|OR_AUTHCFG,
		"LUA Request Check User ID Handle Function "
		"(`lua function has been defined', `user_id, auth_type access_check(uri)')"),

		
#ifdef APR_HAS_THREADS
	AP_INIT_TAKE1("LUA_process_queue_Length", lua_cmd_process_queue_Length,
		NULL, RSRC_CONF,
		"LUA Request Queue Max Length "
		"(`N' - Max numbers of Request in Process Queue)"),
#endif
	AP_INIT_TAKE1("LUA_INIT", lua_cmd_Init,
		NULL, RSRC_CONF,
		"LUA VM Script INIT "
		"(`@PATH --LuaScript FilePath', `lua script content')"),

	AP_INIT_TAKE1("HTTP_INIT", lua_cmd_Http,
		NULL, RSRC_CONF,
		"LUA VM Script HTTP INIT "
		"(`@PATH --LuaScript FilePath', `lua script content')"),

	AP_INIT_TAKE1("LUA_MAIN", lua_cmd_Main,
		NULL, RSRC_CONF,
		"LUA VM Main Script "
		"(`@PATH --LuaScript Main Script FilePath', `lua main script content')"),
	
	AP_INIT_TAKE1("LUA_HANDLE", lua_cmd_Handle,
		NULL, ACCESS_CONF,
		"LUA VM Handle Script "
		"(`@PATH --LuaScript handle Script FilePath', `lua handle script content')"),
	{NULL}
};


static void register_hooks(apr_pool_t *p)
{
	ap_hook_pre_config(modlua_prepare,NULL,NULL,APR_HOOK_MIDDLE);
	ap_hook_post_config(modlua_startup, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_child_init(modlua_init_Child, NULL, NULL, APR_HOOK_MIDDLE); 
	ap_hook_handler(lua_handler, NULL, NULL, APR_HOOK_FIRST);
	ap_hook_access_checker(lua_access, NULL, NULL, APR_HOOK_FIRST);
	ap_hook_check_user_id(lua_check_user_id, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA lua_module =
{
	STANDARD20_MODULE_STUFF,
	create_ml_directory,                        /* dir config creater */
	merge_ml_directory,                        /* dir merger --- default is to override */
	storage_config_server_create,   /* create per-server config structures */
	storage_config_server_merge,    /* merge  per-server config structures */
	lua_cmds,                    /* command apr_table_t */
	register_hooks               /* register hooks */
};


