#ifndef MOD_LUA_H
#define MOD_LUA_H
/*
* Power up our brain...
*/

/* Apache headers */
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_connection.h"
#include "http_request.h"
#include "http_protocol.h"
#include "util_script.h"
#include "util_filter.h"

/* ARP headers */
#include "mpm.h"
#include "apr.h"
#include "apr_strings.h"
#define APR_WANT_STRFUNC
#include "apr_want.h"
#include "apr_tables.h"
#include "apr_lib.h"
#include "apr_fnmatch.h"
#include "apr_strings.h"
#include "apr_dbm.h"
#include "apr_rmm.h"
#include "apr_shm.h"
#include "apr_global_mutex.h"
#include "apr_optional.h"
#include "apr_queue.h"
#include "apr_strings.h"
#include "apr_env.h"
#include "apr_thread_mutex.h"
#include "apr_thread_cond.h"

/* lua headers */
#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

#if 0
#include "ldo.h"
#include "lfunc.h"
#include "lmem.h"
#include "lobject.h"
#include "lopcodes.h"
#include "lstring.h"
#include "lundump.h"
#endif

/* The #ifdef macros are only defined AFTER including the above
* therefore we cannot include these system files at the top  :-(
*/
#if APR_HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#if APR_HAVE_UNISTD_H
#include <unistd.h> /* needed for STDIN_FILENO et.al., at least on FreeBSD */
#endif

#include "storage_util_table.h"
/*
* Provide reasonable default for some defines
*/
#ifndef FALSE
#define FALSE (0)
#endif
#ifndef TRUE
#define TRUE (!FALSE)
#endif
#ifndef PFALSE
#define PFALSE ((void *)FALSE)
#endif
#ifndef PTRUE
#define PTRUE ((void *)TRUE)
#endif
#ifndef UNSET
#define UNSET (-1)
#endif
#ifndef NUL
#define NUL '\0'
#endif
#ifndef RAND_MAX
#include <limits.h>
#define RAND_MAX INT_MAX
#endif

#define APR_SHM_MAXSIZE (64 * 1024 * 1024)

typedef enum {
	STORAGE_SCMODE_UNSET = UNSET,
	STORAGE_SCMODE_NONE  = 0,
	STORAGE_SCMODE_DBM   = 1,
	STORAGE_SCMODE_SHMHT = 2,
	STORAGE_SCMODE_SHMCB = 3,
} storage_scmode_t;

/*
* Define the STORAGE mutex modes
*/
typedef enum {
	STORAGE_MUTEXMODE_UNSET  = UNSET,
	STORAGE_MUTEXMODE_NONE   = 0,
	STORAGE_MUTEXMODE_USED   = 1
} storage_mutexmode_t;

typedef struct {
	pid_t           pid;
	apr_pool_t     *pPool;
	int            bFixed;
	int             nStorageMode;
	char           *szStorageDataFile;
	int             nStorageDataSize;
	apr_shm_t      *pStorageDataMM;
	apr_rmm_t      *pStorageDataRMM;
	apr_table_t    *tStorageDataTable;
	storage_mutexmode_t  nMutexMode;
	apr_lockmech_e  nMutexMech;
	const char     *szMutexFile;
	apr_global_mutex_t   *pMutex;
} LUAConfigRecord;

typedef struct {
	LUAConfigRecord  *mc;
	int                  enabled;
	int                  release;
	int                  storage_timeout;
	char				*init_script;
	char				*main_script;
	char				*http_init;
	char				*access_check;
	char				*check_user_id;
#ifdef APR_HAS_THREADS
	int					 is_threaded_mpm;
	int                  process_queue_length;
	union {
		apr_queue_t			*process_queue;
		lua_State			*L;
	};

#else
	lua_State			*L;
#endif
}LUASrvConfigRec;

#define MAX_HANDLERS 16
/*
** Directory configuration structure.
** Stores a list of valid handlers.
*/
typedef struct {
	char      *handlers;
} ml_directory;

#ifndef BOOL
#define BOOL unsigned int
#endif
#ifndef UCHAR
#define UCHAR unsigned char
#endif

/*  API glue structures  */
extern module AP_MODULE_DECLARE_DATA lua_module;
extern const char lua_valid_mutex_string[];

#define ThisSrvConfig(srv) (LUASrvConfigRec *)ap_get_module_config(srv->module_config,  &lua_module)
#define ThisConfigRecord(srv) (ThisSrvConfig((srv)))->mc
#define VHOST(r) (r->server)
#define RHOST(r) (r->connection->server)
#define SrvConfig4Request(r) ((ThisSrvConfig(VHOST(r))))


#endif
